package com.example.demo

import io.kotest.matchers.collections.shouldHaveSize
import org.junit.jupiter.api.AfterEach
import org.junit.jupiter.api.BeforeEach
import org.junit.jupiter.api.Test
import org.springframework.beans.factory.annotation.Autowired
import org.springframework.boot.test.context.SpringBootTest
import org.springframework.boot.test.context.SpringBootTest.WebEnvironment.RANDOM_PORT
import org.springframework.data.mongodb.core.ReactiveMongoTemplate
import org.springframework.http.HttpHeaders
import org.springframework.http.MediaType
import org.springframework.test.web.reactive.server.WebTestClient
import reactor.core.publisher.Mono
import reactor.test.StepVerifier

@SpringBootTest(webEnvironment = RANDOM_PORT)
class IntegrationTest {

    @Autowired
    lateinit var reactiveMongoTemplate: ReactiveMongoTemplate

    @Autowired
    lateinit var webTestClient: WebTestClient

    @Autowired
    lateinit var anyDocumentRepository: AnyDocumentRepository

    @Autowired
    private lateinit var mongoValidationSetupUtils: MongoValidationSetupUtils

    @BeforeEach
    fun setup() {
        if (reactiveMongoTemplate.collectionExists("any-document").block() != true) {
            reactiveMongoTemplate.createCollection("any-document").block()
        }

        anyDocumentRepository.deleteAll().block()
        mongoValidationSetupUtils.resetValidationRule().block()
    }

    @AfterEach
    fun cleanUp() {
        anyDocumentRepository.deleteAll().block()
        mongoValidationSetupUtils.resetValidationRule().block()
    }

    @Test
    fun `should rollback AnyDocument save on exception`() {
        // given
        mongoValidationSetupUtils.setupValidationRule("2").block()

        val request = AnyDocumentRequest(
            list = listOf(
                AnyDocumentDTO("1"),
                AnyDocumentDTO("2")
            )
        )

        // when
        (1..100).map {
            webTestClient
                .post()
                .uri { it.path("/foo").build() }
                .body(Mono.just(request), AnyDocumentRequest::class.java)
                .header(HttpHeaders.ACCEPT, MediaType.APPLICATION_JSON_VALUE)
                .exchange()
                .expectStatus().is5xxServerError
        }

        // then
        val storedDocuments = anyDocumentRepository.findAll()

        StepVerifier.create(storedDocuments)
            .expectNextCount(0)
            .verifyComplete()
    }

    @Test
    fun `should save AnyDocuments transactionally`() {
        // given
        val request = AnyDocumentRequest(
            list = listOf(
                AnyDocumentDTO("1"),
                AnyDocumentDTO("2")
            )
        )

        // when
        val responseSpec = webTestClient
            .post()
            .uri { it.path("/foo").build() }
            .body(Mono.just(request), AnyDocumentRequest::class.java)
            .header(HttpHeaders.ACCEPT, MediaType.APPLICATION_JSON_VALUE)
            .exchange()

        val storedDocuments = anyDocumentRepository
            .findAll()
            .collectList()
            .block()

        // then
        responseSpec.expectStatus().is2xxSuccessful
        storedDocuments!!.shouldHaveSize(2)
    }

}
